// -*- compile-command: "make run" -*-
/*
 * Copyright (C) 2024
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "p1.h"
#include <iostream>
#include <cstdint>
#include <iomanip>
#include <cmath>
#include <limits>

// -- Mapping functions: --------------------------------------------------
etype_t power2(etype_t e) { return e * e; }
etype_t sum10(etype_t e)  { return e + 10; }
etype_t triple(etype_t e) { return e * 3; }

// -- Tests: --------------------------------------------------------------
void test1() {
  std::cout << "\nTest1\n";

  // [[5,3],[2,1]]
  uint32_t R = 2;
  uint32_t C = 2;
  Matrix m = mreserve(R, C);

  m.data[0][0] = 5;
  m.data[0][1] = 3;
  m.data[1][0] = 2;
  m.data[1][1] = 1;

  std::cout << "Original m:\n";
  mshow(m);

  map_elements(m, power2);

  std::cout << "\nMapped (power2) m:\n";
  mshow(m);

  mfree(m);

  cout << "\n--------------------------" << endl;
}

void test2() {
  std::cout << "\nTest2\n";

  // [[1,2,3],[4,5,6],[7,8,9]]
  uint32_t R = 3;
  uint32_t C = 3;
  Matrix m = mreserve(R, C);

  etype_t e = 1;
  for (auto i = 0u; i < m.rows; i++)
    for (auto j = 0u; j < m.cols; j++) {
      m.data[i][j] = e;
      e += 1;
    }

  std::cout << "Original m:\n";
  mshow(m);

  map_elements(m, sum10);

  std::cout << "\nMapped (sum10) m:\n";
  mshow(m);

  mfree(m);

  cout << "\n--------------------------" << endl;
}

void test3() {
  std::cout << "\nTest3\n";

  // [[1,2,3,8],[4,5,6,2],[1,7,8,9],[3,4,18,29]]
  uint32_t R = 4;
  uint32_t C = 4;
  Matrix m = mreserve(R, C);

  // row 0
  m.data[0][0] = 1;
  m.data[0][1] = 2;
  m.data[0][2] = 3;
  m.data[0][3] = 8;

  // row 1
  m.data[1][0] = 4;
  m.data[1][1] = 5;
  m.data[1][2] = 6;
  m.data[1][3] = 2;

  // row 2
  m.data[2][0] = 1;
  m.data[2][1] = 7;
  m.data[2][2] = 8;
  m.data[2][3] = 9;

  // row 3
  m.data[3][0] = 3;
  m.data[3][1] = 4;
  m.data[3][2] = 18;
  m.data[3][3] = 29;

  std::cout << "Original m:\n";
  mshow(m);

  map_elements(m, triple);

  std::cout << "\nMapped (triple) m:\n";
  mshow(m);

  mfree(m);
  cout << "\n--------------------------" << endl;
}

void test4() {
  std::cout << "\nTest4\n";

  // [[5,3],[2,1]]
  uint32_t R = 2;
  uint32_t C = 2;
  Matrix m = mreserve(R, C);

  m.data[0][0] = 5;
  m.data[0][1] = 3;
  m.data[1][0] = 2;
  m.data[1][1] = 1;

  auto nm = matrix_grow(m, 2, 1);

  std::cout << "original is r: " << m.rows << " , c: " << m.cols << '\n';
  mshow(m);
  std::cout << "\nGrow 2u:\nnew is r: " << nm.rows << " , c: " << nm.cols << '\n';
  mshow(nm);

  mfree(m);
  mfree(nm);
  cout << "\n--------------------------" << endl;
}

void test5() {
  std::cout << "\nTest5\n";

  // [[1,2,3,8],[4,5,6,2],[1,7,8,9],[3,4,18,29]]
  uint32_t R = 4;
  uint32_t C = 4;
  Matrix m = mreserve(R, C);

  // row 0
  m.data[0][0] = 1;
  m.data[0][1] = 2;
  m.data[0][2] = 3;
  m.data[0][3] = 8;

  // row 1
  m.data[1][0] = 4;
  m.data[1][1] = 5;
  m.data[1][2] = 6;
  m.data[1][3] = 2;

  // row 2
  m.data[2][0] = 1;
  m.data[2][1] = 7;
  m.data[2][2] = 8;
  m.data[2][3] = 9;

  // row 3
  m.data[3][0] = 3;
  m.data[3][1] = 4;
  m.data[3][2] = 18;
  m.data[3][3] = 29;

  std::cout << "Original m:\n";
  mshow(m);

  auto nm = matrix_shrink(m, 2);
  if (nm.data == nullptr)
    std::cout << "Couldn't shrink by 2 units.\n";

  mfree(m);
  cout << "\n--------------------------" << endl;
}

void test6() {
  std::cout << "\nTest6\n";

  // [[1,2,3,8],[4,5,6,2],[1,7,8,9],[3,4,18,29]]
  uint32_t R = 4;
  uint32_t C = 4;
  Matrix m = mreserve(R, C);

  // row 0
  m.data[0][0] = 1;
  m.data[0][1] = 2;
  m.data[0][2] = 3;
  m.data[0][3] = 8;

  // row 1
  m.data[1][0] = 4;
  m.data[1][1] = 5;
  m.data[1][2] = 6;
  m.data[1][3] = 2;

  // row 2
  m.data[2][0] = 1;
  m.data[2][1] = 7;
  m.data[2][2] = 8;
  m.data[2][3] = 9;

  // row 3
  m.data[3][0] = 3;
  m.data[3][1] = 4;
  m.data[3][2] = 18;
  m.data[3][3] = 29;

  std::cout << "Original m:\n";
  mshow(m);

  auto nm = matrix_shrink(m, 1);
  if (nm.data == nullptr)
    std::cout << "Couldn't shrink by 1 unit.\n";
  else {
    std::cout << "\nShrink 1u:\nnew is r: " << nm.rows << " , c: " << nm.cols << '\n';
    mshow(nm);
    mfree(nm);
  }

  mfree(m);
  cout << "\n--------------------------" << endl;
}

void test7() {
  std::cout << "\nTest7\n";

  // [[1,2,3],[4,5,6],[1,7,8],[3,4,18]]
  uint32_t R = 4;
  uint32_t C = 3;
  Matrix m = mreserve(R, C);

  // row 0
  m.data[0][0] = 1;
  m.data[0][1] = 2;
  m.data[0][2] = 3;

  // row 1
  m.data[1][0] = 4;
  m.data[1][1] = 5;
  m.data[1][2] = 6;

  // row 2
  m.data[2][0] = 1;
  m.data[2][1] = 7;
  m.data[2][2] = 8;

  // row 3
  m.data[3][0] = 3;
  m.data[3][1] = 4;
  m.data[3][2] = 18;

  std::cout << "Original m:\n";
  mshow(m);

  auto nm = matrix_shrink(m, 2);
  if (nm.data == nullptr)
    std::cout << "Couldn't shrink by 2 units.\n";
  else {
    std::cout << "\nShrink 2u:\nnew is r: " << nm.rows << " , c: " << nm.cols << '\n';
    mshow(nm);
    mfree(nm);
  }

  nm = matrix_shrink(m, 1);
  if (nm.data == nullptr)
    std::cout << "Couldn't shrink by 1 unit.\n";
  else {
    std::cout << "\nShrink 1u:\nnew is r: " << nm.rows << " , c: " << nm.cols << '\n';
    mshow(nm);
    mfree(nm);
  }

  mfree(m);
  cout << "\n--------------------------" << endl;
}

// -- Main program: -------------------------------------------------------
int main()
{
  cout << "Ejercicio 1\n===========\n";
  test1();
  test2();
  test3();
  cout << "\nEjercicio 2\n===========\n";
  test4();
  test5();
  test6();
  test7();

  return 0;
}
